@@ -20,6 +20,7 @@ import io
 import json
 import os
 import pycurl
+import re
 import six
 import sys
 import threading
@@ -39,7 +40,24 @@ class Error(Exception):
     General exception which is thrown by SDK,
     indicates that some exception happened in SDK.
     """
-    pass
+
+    def __init__(self, message, code=None, fault=None):
+        """
+        Creates an instance of Error class.
+
+        `message`:: The exception message.
+
+        `code`:: An error code associated to the error. For HTTP related
+        errors, this will be the HTTP response code returned by the server.
+        For example, if retrieving of a virtual machine fails because it
+        doesn't exist this attribute will contain the integer value 404. Note
+        that this may be `nil` if the error is not HTTP related.
+
+        `fault`:: The `Fault` object associated to the error.
+        """
+        super(Error, self).__init__(message)
+        self.code = code
+        self.fault = fault
 
 
 class List(list):
@@ -119,6 +137,34 @@ class Connection(object):
     services provided by the API.
     """
 
+    # Regular expression used to check XML content type.
+    __XML_CONTENT_TYPE_RE = re.compile("^\s*(application|text)/xml\s*(;.*)?$")
+
+    # Regular expression used to check JSON content type.
+    __JSON_CONTENT_TYPE_RE = re.compile("^\s*(application|text)/json\s*(;.*)?$")
+
+    # The typical URL path, used just to generate informative error messages.
+    __TYPICAL_PATH = '/ovirt-engine/api'
+
+    # Debug types that we know how to handle. Everything else will be
+    # silently ignored.
+    __KNOWN_DEBUG_TYPES = [
+        pycurl.INFOTYPE_TEXT,
+        pycurl.INFOTYPE_HEADER_IN,
+        pycurl.INFOTYPE_HEADER_OUT,
+        pycurl.INFOTYPE_DATA_IN,
+        pycurl.INFOTYPE_DATA_OUT,
+    ]
+
+    # Prefixes to use for debug data types:
+    __DEBUG_PREFIXES = {
+        pycurl.INFOTYPE_TEXT: '* ',
+        pycurl.INFOTYPE_HEADER_IN: '> ',
+        pycurl.INFOTYPE_HEADER_OUT: '< ',
+        pycurl.INFOTYPE_DATA_IN: '> ',
+        pycurl.INFOTYPE_DATA_OUT: '< ',
+    }
+
     def __init__(
         self,
         url=None,
@@ -131,10 +177,13 @@ class Connection(object):
         log=None,
         kerberos=False,
         timeout=0,
-        compress=False,
+        compress=True,
         sso_url=None,
         sso_revoke_url=None,
         sso_token_name='access_token',
+        headers=None,
+        pipeline=0,
+        connections=0,
     ):
         """
         Creates a new connection to the API server.
@@ -178,9 +227,11 @@ class Connection(object):
         will be raised.
 
         `compress`:: A boolean flag indicating if the SDK should ask
-        the server to send compressed responses. The default is `False`.
+        the server to send compressed responses. The default is `True`.
         Note that this is a hint for the server, and that it may return
         uncompressed data even when this parameter is set to `True`.
+        Note that compression will be disabled if user pass `debug`
+        parameter set to `true`, so the debug messages are in plain text.
 
         `sso_url`:: A string containing the base SSO URL of the serve.
         Default SSO url is computed from the `url` if no `sso_url` is provided.
@@ -194,6 +245,17 @@ class Connection(object):
 
         `sso_token_name`:: The token name in the JSON SSO response returned
         from the SSO server. Default value is `access_token`.
+
+        `headers`:: A dictionary with headers which should be send with every
+        request.
+
+        `connections`:: The maximum number of connections to open to the host.
+        If the value is `0` (the default) then the number of connections will
+        be unlimited.
+
+        `pipeline`:: The maximum number of request to put in an HTTP pipeline
+        without waiting for the response. If the value is `0` (the default)
+        then pipelining is disabled.
         """
 
         # Check mandatory parameters:
@@ -215,6 +277,11 @@ class Connection(object):
         self._password = password
         self._sso_token = token
         self._kerberos = kerberos
+        self._ca_file = ca_file
+        self._insecure = insecure
+        self._timeout = timeout
+        self._debug = debug
+        self._compress = compress
 
         # The curl object can be used by several threads, but not
         # simultaneously, so we need a lock to prevent that:
@@ -225,31 +292,19 @@ class Connection(object):
         self._sso_revoke_url = sso_revoke_url
         self._sso_token_name = sso_token_name
 
+        # Headers:
+        self._headers = headers or {}
+
         # Create the curl handle that manages the pool of connections:
-        self._curl = pycurl.Curl()
-        self._curl.setopt(pycurl.COOKIEFILE, '/dev/null')
-        self._curl.setopt(pycurl.COOKIEJAR, '/dev/null')
+        self._multi = pycurl.CurlMulti()
+        self._multi.setopt(pycurl.M_PIPELINING, bool(pipeline))
+        # Since libcurl 7.30.0:
+        if hasattr(pycurl, 'M_MAX_PIPELINE_LENGTH'):
+            self._multi.setopt(pycurl.M_MAX_PIPELINE_LENGTH, pipeline)
+            self._multi.setopt(pycurl.M_MAX_HOST_CONNECTIONS, connections)
 
-        # Configure TLS parameters:
-        if url.startswith('https'):
-            self._curl.setopt(pycurl.SSL_VERIFYPEER, 0 if insecure else 1)
-            self._curl.setopt(pycurl.SSL_VERIFYHOST, 0 if insecure else 2)
-            if ca_file is not None:
-                self._curl.setopt(pycurl.CAINFO, ca_file)
-
-        # Configure timeouts:
-        self._curl.setopt(pycurl.TIMEOUT, timeout)
-
-        # Configure compression of responses (setting the value to a zero
-        # length string means accepting all the compression types that
-        # libcurl supports):
-        if compress:
-            self._curl.setopt(pycurl.ENCODING, '')
-
-        # Configure debug mode:
-        if debug and log is not None:
-            self._curl.setopt(pycurl.VERBOSE, 1)
-            self._curl.setopt(pycurl.DEBUGFUNCTION, self._curl_debug)
+        # Connections:
+        self._curls = set()
 
         # Initialize the reference to the system service:
         self.__system_service = None
@@ -297,15 +352,41 @@ class Connection(object):
         if self._sso_token is None:
             self._sso_token = self._get_access_token()
 
+        # Init curl easy:
+        curl = pycurl.Curl()
+        curl.setopt(pycurl.COOKIEFILE, '/dev/null')
+        curl.setopt(pycurl.COOKIEJAR, '/dev/null')
+
+        # Configure TLS parameters:
+        if self._url.startswith('https'):
+            curl.setopt(pycurl.SSL_VERIFYPEER, 0 if self._insecure else 1)
+            curl.setopt(pycurl.SSL_VERIFYHOST, 0 if self._insecure else 2)
+            if self._ca_file is not None:
+                curl.setopt(pycurl.CAINFO, self._ca_file)
+
+        # Configure timeouts:
+        curl.setopt(pycurl.TIMEOUT, self._timeout)
+
+        # Configure compression of responses (setting the value to a zero
+        # length string means accepting all the compression types that
+        # libcurl supports):
+        if self._compress and not self._debug:
+            curl.setopt(pycurl.ENCODING, '')
+
+        # Configure debug mode:
+        if self._debug and self._log is not None:
+            curl.setopt(pycurl.VERBOSE, 1)
+            curl.setopt(pycurl.DEBUGFUNCTION, self._curl_debug)
+
         # Set the method:
-        self._curl.setopt(pycurl.CUSTOMREQUEST, request.method)
+        curl.setopt(pycurl.CUSTOMREQUEST, request.method)
 
         # Build the URL:
         url = self._build_url(
             path=request.path,
             query=request.query,
         )
-        self._curl.setopt(pycurl.URL, url)
+        curl.setopt(pycurl.URL, url)
 
         # Older versions of the engine (before 4.1) required the
         # 'all_content' parameter as an HTTP header instead of a query
@@ -317,11 +398,18 @@ class Connection(object):
             if all_content is not None:
                 request.headers['All-Content'] = all_content
 
+        # Add global headers:
+        headers_dict = self._headers.copy()
+
         # Add headers, avoiding those that have no value:
         header_lines = []
         for header_name, header_value in request.headers.items():
             if header_value is not None:
-                header_lines.append('%s: %s' % (header_name, header_value))
+                headers_dict[header_name] = header_value
+
+        for header_name, header_value in headers_dict.items():
+            header_lines.append('%s: %s' % (header_name, header_value))
+
         header_lines.append('User-Agent: PythonSDK/%s' % version.VERSION)
         header_lines.append('Version: 4')
         header_lines.append('Content-Type: application/xml')
@@ -345,40 +433,65 @@ class Connection(object):
                 )
 
         # Copy headers and the request body to the curl object:
-        self._curl.setopt(pycurl.HTTPHEADER, header_lines)
+        curl.setopt(pycurl.HTTPHEADER, header_lines)
         body = request.body
         if body is None:
             body = ''
-        self._curl.setopt(pycurl.COPYPOSTFIELDS, body)
+
+        # HTTP pipelining is valid only for idempotent operations,
+        # pycurl automatically disables pipelining if COPYPOSTFIELDS
+        # is set, even if pipelining is explicitly set.
+        if request.method in ['POST', 'PUT']:
+            curl.setopt(pycurl.COPYPOSTFIELDS, body.encode('utf-8'))
 
         # Prepare the buffers to receive the response:
         body_buf = io.BytesIO()
         headers_buf = io.BytesIO()
-        self._curl.setopt(pycurl.WRITEFUNCTION, body_buf.write)
-        self._curl.setopt(pycurl.HEADERFUNCTION, headers_buf.write)
+        curl.setopt(pycurl.WRITEFUNCTION, body_buf.write)
+        curl.setopt(pycurl.HEADERFUNCTION, headers_buf.write)
 
-        # Send the request and wait for the response:
-        self._curl.perform()
+        # Add the curl easy to the multi handle:
+        self._multi.add_handle(curl)
 
-        # Extract the response code and body:
-        response = Response()
-        response.code = self._curl.getinfo(pycurl.HTTP_CODE)
-        response.body = body_buf.getvalue()
+        return curl, body_buf, headers_buf, request
 
-        # The response code can be extracted directly, but cURL doesn't
-        # have a method to extract the response message, so we have to
-        # parse the first header line to find it:
-        response.reason = ""
-        headers_text = headers_buf.getvalue().decode('ascii')
-        header_lines = headers_text.split('\n')
-        if len(header_lines) >= 1:
-            response_line = header_lines[0]
-            response_fields = response_line.split()
-            if len(response_fields) >= 3:
-                response.reason = ' '.join(response_fields[2:])
+    def wait(self, context, failed_auth=False):
+        with self._curl_lock:
+            return self.__wait(context, failed_auth)
 
-        # Return the response:
-        return response
+    def __wait(self, context, failed_auth=False):
+        while True:
+            while True:
+                ret, _ = self._multi.perform()
+                if ret != pycurl.E_CALL_MULTI_PERFORM:
+                    break
+            while True:
+                num_q, ok_list, err_list = self._multi.info_read()
+                self._curls = self._curls.union(set(ok_list))
+                if err_list:
+                    raise Error("Failed to read response.")
+                elif context[0] in self._curls:
+                    # Remove the curl:
+                    self._curls.remove(context[0])
+                    self._multi.remove_handle(context[0])
+
+                    # Read the response:
+                    response = self._read_reponse(context)
+
+                    # If the request failed because of authentication, and it
+                    # wasn't a request to the SSO service, then the most likely
+                    # cause is an expired SSO token. In this case we need to
+                    # request a new token, and try the original request again, but
+                    # only once. It if fails again, we just return the failed
+                    # response.
+                    if response.code == 401 and not failed_auth:
+                        self._sso_token = self._get_access_token()
+                        context = self.__send(context[3])
+                        response = self.__wait(context, True)
+                    return response
+                elif num_q == 0:
+                    break
+            self._multi.select(1.0)
 
     @property
     def url(self):
@@ -461,11 +574,6 @@ class Connection(object):
                 'password': self._password,
             })
 
-        # Set proper Authorization header if using kerberos:
-        if self._kerberos:
-            self._curl.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_GSSNEGOTIATE)
-            self._curl.setopt(pycurl.USERPWD, ':')
-
         # Send SSO request:
         sso_response = self._get_sso_response(self._sso_url, post_data)
 
@@ -488,23 +596,62 @@ class Connection(object):
         """
 
         # Set HTTP method and URL:
-        self._curl.setopt(pycurl.CUSTOMREQUEST, 'POST')
-        self._curl.setopt(pycurl.URL, url)
+        curl = pycurl.Curl()
+        curl.setopt(pycurl.CUSTOMREQUEST, 'POST')
+        curl.setopt(pycurl.URL, url)
+
+        # Set proper Authorization header if using kerberos:
+        if self._kerberos:
+            curl.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_GSSNEGOTIATE)
+            curl.setopt(pycurl.USERPWD, ':')
+
+        # Configure debug mode:
+        if self._debug and self._log is not None:
+            curl.setopt(pycurl.VERBOSE, 1)
+            curl.setopt(pycurl.DEBUGFUNCTION, self._curl_debug)
+
+        # Configure TLS parameters:
+        if self._url.startswith('https'):
+            curl.setopt(pycurl.SSL_VERIFYPEER, 0 if self._insecure else 1)
+            curl.setopt(pycurl.SSL_VERIFYHOST, 0 if self._insecure else 2)
+            if self._ca_file is not None:
+                curl.setopt(pycurl.CAINFO, self._ca_file)
 
         # Prepare headers:
         header_lines = [
             'User-Agent: PythonSDK/%s' % version.VERSION,
             'Accept: application/json'
         ]
-        self._curl.setopt(pycurl.HTTPHEADER, header_lines)
-        self._curl.setopt(pycurl.COPYPOSTFIELDS, urlencode(params))
+        curl.setopt(pycurl.HTTPHEADER, header_lines)
+        curl.setopt(pycurl.COPYPOSTFIELDS, urlencode(params))
 
         # Prepare the buffer to receive the response:
         body_buf = io.BytesIO()
-        self._curl.setopt(pycurl.WRITEFUNCTION, body_buf.write)
+        headers_buf = io.BytesIO()
+
+        # We ignore all headers which are received before the last response
+        # from the server, because we can be forwarded by apache, and we
+        # are interested only in last response:
+        def write_header(buf):
+            if buf.startswith(b'HTTP/'):
+                headers_buf.truncate(0)
+                headers_buf.seek(0)
+            headers_buf.write(buf)
+
+        curl.setopt(pycurl.WRITEFUNCTION, body_buf.write)
+        curl.setopt(pycurl.HEADERFUNCTION, write_header)
 
         # Send the request and wait for the response:
-        self._curl.perform()
+        curl.perform()
+        curl.close()
+
+        # Get headers:
+        headers_text = headers_buf.getvalue().decode('ascii')
+        header_lines = headers_text.split('\n')
+
+        # Check the returned content type:
+        self._check_content_type(self.__JSON_CONTENT_TYPE_RE, 'JSON', header_lines)
+
         return json.loads(body_buf.getvalue().decode('utf-8'))
 
     def system_service(self):
@@ -606,7 +753,7 @@ class Connection(object):
 
         # Release resources used by the cURL handle:
         with self._curl_lock:
-            self._curl.close()
+            self._multi.close()
 
     def _build_url(self, path='', query=None):
         """
@@ -635,11 +782,111 @@ class Connection(object):
             url = '%s?%s' % (url, urlencode(sorted(query.items())))
         return url
 
+    def check_xml_content_type(self, response):
+        """
+         Checks that the content type of the given response is XML. If it is
+         XML then it does nothing. If it isn't XML then it raises an
+         exception.
+
+         `response`:: The HTTP response to check.
+        """
+        return self._check_content_type(
+            self.__XML_CONTENT_TYPE_RE,
+            'XML',
+            response.headers
+        )
+
+    def check_json_content_type(self, response):
+        """
+         Checks that the content type of the given response is JSON. If it is
+         JSON then it does nothing. If it isn't JSON then it raises an
+         exception.
+
+         `response`:: The HTTP response to check.
+        """
+        return self._check_content_type(
+            self.__JSON_CONTENT_TYPE_RE,
+            'JSON',
+            response.headers
+        )
+
+    def _check_content_type(self, expected_re, expected_name, headers):
+        """
+        Checks the given content type and raises an exception if it isn't the
+        expected one.
+
+        `expected_re`:: The regular expression used to check the expected
+                        content type.
+        `expected_name`:: The name of the expected content type.
+        `headers`:: The HTTP headers to check.
+        """
+        content_type = self._get_header_value(headers, 'content-type')
+        if expected_re.match(content_type) is None:
+            msg = "The response content type '{}' isn't the expected {}".format(
+                content_type,
+                expected_name,
+            )
+            url = urlparse(self._url)
+            if url.path != self.__TYPICAL_PATH:
+                msg += (
+                    ". Is the path '{}' included in the 'url' "
+                    "parameter correct?"
+                ).format(url.path)
+                msg += " The typical one is '{}'".format(self.__TYPICAL_PATH)
+            raise Error(msg)
+
+    def _read_reponse(self, context):
+        """
+        Read the response.
+
+        `context`:: tuple which contains cur easy, response body,
+        response headers, original request
+        """
+        # Extract the response code and body:
+        response = Response()
+        response.code = context[0].getinfo(pycurl.HTTP_CODE)
+        response.body = context[1].getvalue()
+
+        # The response code can be extracted directly, but cURL doesn't
+        # have a method to extract the response message, so we have to
+        # parse the first header line to find it:
+        response.reason = ""
+        headers_text = context[2].getvalue().decode('ascii')
+        header_lines = headers_text.split('\n')
+        response.headers = header_lines
+        if len(header_lines) >= 1:
+            response_line = header_lines[0]
+            response_fields = response_line.split()
+            if len(response_fields) >= 3:
+                response.reason = ' '.join(response_fields[2:])
+
+        context[0].close()
+        # Return the response:
+        return response
+
+    def _get_header_value(self, headers, name):
+        """
+        Return header value by its name.
+
+        `headers`:: list of headers
+        `name`:: name of the header
+        """
+        return next(
+            (h.split(':')[1].strip() for h in headers if h.lower().startswith(name)),
+            None
+        )
+
     def _curl_debug(self, debug_type, data):
         """
         This is the implementation of the cURL debug callback.
         """
 
+        # Exclude all types of debug data that we don't know how to
+        # handle, as trying to decode and manipulate that data as
+        # strings will likely fail.
+        if debug_type not in self.__KNOWN_DEBUG_TYPES:
+            return
+
         # Some versions of PycURL provide the debug data as strings, and
         # some as arrays of bytes, so we need to check the type of the
         # provided data and convert it to strings before trying to
@@ -649,7 +896,10 @@ class Connection(object):
         # Split the debug data into lines and send a debug message for
         # each line:
         lines = text.replace('\r\n', '\n').strip().split('\n')
+        prefix = self.__DEBUG_PREFIXES.get(debug_type)
         for line in lines:
+            if prefix is not None:
+                line = prefix + line
             self._log.debug(line)
 
     def __enter__(self):
@@ -702,3 +952,17 @@ class ConnectionBuilder(object):
         """
 
         return Connection(**self._kwargs)
+
+
+# We need to import readers and writers here, so the generic
+# writer and reader are initialized. This has also benefit that
+# the readers/writers/types/services can be used right after the
+# top level sdk module import, for example:
+#
+#   import ovirtsdk4 as sdk
+#   vm = sdk.types.Vm()
+#
+from ovirtsdk4 import readers  # noqa: F401
+from ovirtsdk4 import services  # noqa: F401
+from ovirtsdk4 import types  # noqa: F401
+from ovirtsdk4 import writers  # noqa: F401
